package ru.dbotthepony.mc.otm.block.entity

import com.google.common.collect.ImmutableList
import com.mojang.serialization.Codec
import net.minecraft.core.BlockPos
import net.minecraft.core.HolderLookup
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.ListTag
import net.minecraft.network.chat.Component
import net.minecraft.world.item.ItemStack
import net.minecraft.world.item.TooltipFlag
import net.minecraft.world.level.BlockGetter
import net.minecraft.world.level.block.Block
import net.minecraft.world.level.block.entity.BlockEntityType
import net.minecraft.world.level.block.state.BlockState
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.mc.otm.capability.IMatteryUpgrade
import ru.dbotthepony.mc.otm.capability.energy.IMatteryEnergyStorage
import ru.dbotthepony.mc.otm.capability.energy.WorkerEnergyStorage
import ru.dbotthepony.mc.otm.core.immutableList
import ru.dbotthepony.mc.otm.core.nbt.getCompoundList
import ru.dbotthepony.mc.otm.core.nbt.set
import ru.dbotthepony.mc.otm.core.util.countingLazy

/**
 * Simple machine, which can work on only one job type.
 *
 * From technical point, this is a specialized use case of [MachineJobEventLoop].
 */
abstract class MatteryWorkerBlockEntity<JobType : IJob>(
	type: BlockEntityType<*>,
	blockPos: BlockPos,
	blockState: BlockState,
	jobCodec: Codec<JobType>,
	maxJobs: Int = 1,
) : MatteryPoweredBlockEntity(type, blockPos, blockState) {
	val jobEventLoops: ImmutableList<MachineJobEventLoop<JobType>> = immutableList(maxJobs) { id ->
		object : MachineJobEventLoop<JobType>(jobCodec) {
			override val energy: IMatteryEnergyStorage?
				get() = this@MatteryWorkerBlockEntity.energy
			override val isBlockedByRedstone: Boolean
				get() = redstoneControl.isBlockedByRedstone
			override val upgrades: IMatteryUpgrade?
				get() = this@MatteryWorkerBlockEntity.upgrades

			override fun onJobFinish(status: JobStatus<JobType>) {
				return this@MatteryWorkerBlockEntity.onJobFinish(status, id)
			}

			override fun computeNextJob(): JobContainer<JobType> {
				return this@MatteryWorkerBlockEntity.computeNextJob(id)
			}

			override fun jobUpdated(new: JobType?, old: JobType?) {
				this@MatteryWorkerBlockEntity.jobUpdated(new, old, id)
			}

			override fun onJobTick(status: JobStatus<JobType>) {
				super@MatteryWorkerBlockEntity.markDirtyFast()
				return this@MatteryWorkerBlockEntity.onJobTick(status, id)
			}
		}
	}

	open val upgrades: IMatteryUpgrade? get() = null

	var balanceInputs = false

	init {
		savetables.bool(::balanceInputs)
	}

	protected open fun jobUpdated(new: JobType?, old: JobType?, id: Int) {}
	protected abstract fun onJobFinish(status: JobStatus<JobType>, id: Int)
	protected abstract fun computeNextJob(id: Int): JobContainer<JobType>

	protected open fun onJobTick(status: JobStatus<JobType>, id: Int) {}

	override fun saveShared(nbt: CompoundTag, registry: HolderLookup.Provider) {
		super.saveShared(nbt, registry)
		nbt["jobs"] = ListTag().also {
			for ((i, job) in jobEventLoops.withIndex()) {
				it.add(job.serializeNBT(registry).also {
					it["_id"] = i
				})
			}
		}
	}

	override fun loadAdditional(nbt: CompoundTag, registry: HolderLookup.Provider) {
		super.loadAdditional(nbt, registry)

		for (v in nbt.getCompoundList("jobs")) {
			if ("_id" in v) {
				val id = v.getInt("_id")

				if (id in jobEventLoops.indices) {
					jobEventLoops[id].deserializeNBT(registry, v)
				}
			}
		}
	}

	override fun setChanged() {
		super.setChanged()
		jobEventLoops.forEach { it.isIdling = false }
	}

	override fun markDirtyFast() {
		super.markDirtyFast()
		jobEventLoops.forEach { it.isIdling = false }
	}

	protected fun energyLevelUpdated() {
		super.markDirtyFast()
		jobEventLoops.forEach { it.notify(MachineJobEventLoop.IdleReason.POWER) }
	}

	protected fun itemContainerUpdated() {
		super.setChanged()
		jobEventLoops.forEach { it.notify(MachineJobEventLoop.IdleReason.ITEM) }
	}

	protected fun matterLevelUpdated() {
		super.markDirtyFast()
		jobEventLoops.forEach { it.notify(MachineJobEventLoop.IdleReason.MATTER) }
	}

	override fun redstoneStatusUpdated(newBlocked: Boolean, oldBlocked: Boolean) {
		super.redstoneStatusUpdated(newBlocked, oldBlocked)
		jobEventLoops.forEach { it.isIdling = newBlocked }
	}

	private val hasWorkerState by countingLazy(blockStateChangesCounter) {
		this.blockState.hasProperty(WorkerState.WORKER_STATE)
	}

	private val isWorkingState by countingLazy(blockStateChangesCounter) {
		hasWorkerState && this.blockState.getValue(WorkerState.WORKER_STATE) == WorkerState.WORKING
	}

	private val isErrorState by countingLazy(blockStateChangesCounter) {
		hasWorkerState && this.blockState.getValue(WorkerState.WORKER_STATE) == WorkerState.ERROR
	}

	private val isIdleState by countingLazy(blockStateChangesCounter) {
		hasWorkerState && this.blockState.getValue(WorkerState.WORKER_STATE) == WorkerState.IDLE
	}

	override fun tick() {
		super.tick()
		jobEventLoops.forEach { it.think() }

		if (hasWorkerState) {
			if (jobEventLoops.any { it.workingTicksAnim > 20 }) {
				if (!isWorkingState) level?.setBlock(blockPos, blockState.setValue(WorkerState.WORKER_STATE, WorkerState.WORKING), Block.UPDATE_CLIENTS)
			} else if (jobEventLoops.any { it.errorTicksAnim > 20 }) {
				if (!isErrorState) level?.setBlock(blockPos, blockState.setValue(WorkerState.WORKER_STATE, WorkerState.ERROR), Block.UPDATE_CLIENTS)
			} else if (jobEventLoops.all { it.idleTicksAnim > 20 }) {
				if (!isIdleState) level?.setBlock(blockPos, blockState.setValue(WorkerState.WORKER_STATE, WorkerState.IDLE), Block.UPDATE_CLIENTS)
			}
		}
	}
}
