package ru.dbotthepony.mc.otm.item

import net.minecraft.ChatFormatting
import net.minecraft.nbt.CompoundTag
import net.minecraft.network.chat.Component
import net.minecraft.server.level.ServerPlayer
import net.minecraft.world.InteractionHand
import net.minecraft.world.InteractionResultHolder
import net.minecraft.world.effect.MobEffectInstance
import net.minecraft.world.effect.MobEffects
import net.minecraft.world.entity.LivingEntity
import net.minecraft.world.entity.player.Player
import net.minecraft.world.item.*
import net.minecraft.world.level.Level
import net.neoforged.neoforge.capabilities.Capabilities
import net.neoforged.neoforge.capabilities.RegisterCapabilitiesEvent
import ru.dbotthepony.mc.otm.config.BatteryBalanceValues
import ru.dbotthepony.mc.otm.capability.*
import ru.dbotthepony.mc.otm.capability.energy.EnergyCapacitorItem
import ru.dbotthepony.mc.otm.capability.energy.getBarColor
import ru.dbotthepony.mc.otm.capability.energy.getBarWidth
import ru.dbotthepony.mc.otm.client.minecraft
import ru.dbotthepony.mc.otm.config.ItemsConfig
import ru.dbotthepony.mc.otm.core.*
import ru.dbotthepony.mc.otm.core.math.Decimal
import ru.dbotthepony.mc.otm.registry.CapabilitiesRegisterListener
import ru.dbotthepony.mc.otm.registry.MDamageTypes
import ru.dbotthepony.mc.otm.registry.MatteryDamageSource
import ru.dbotthepony.mc.otm.runIfClient
import kotlin.math.roundToInt

open class BatteryItem : MatteryItem, CapabilitiesRegisterListener {
	val _capacity: () -> Decimal
	val _receive: () -> Decimal
	val _extract: () -> Decimal
	val _initialBatteryLevel: () -> Decimal

	val capacity: Decimal
		get() = _capacity.invoke()
	val receive: Decimal
		get() = _receive.invoke()
	val extract: Decimal
		get() = _extract.invoke()
	val initialBatteryLevel: Decimal
		get() = _initialBatteryLevel.invoke()

	constructor(
		storage: Decimal,
		receive: Decimal,
		extract: Decimal = receive,
		initialBatteryLevel: Decimal = Decimal.ZERO
	) : super(Properties().stacksTo(1)) {
		this._capacity = { storage }
		this._receive = { receive }
		this._extract = { extract }
		this._initialBatteryLevel = { initialBatteryLevel }
	}

	constructor(
		storage: () -> Decimal,
		receive: () -> Decimal,
		extract: () -> Decimal = receive,
		initialBatteryLevel: () -> Decimal = { Decimal.ZERO }
	) : super(Properties().stacksTo(1)) {
		this._capacity = storage
		this._receive = receive
		this._extract = extract
		this._initialBatteryLevel = initialBatteryLevel
	}

	constructor(values: BatteryBalanceValues) : super(Properties().stacksTo(1)) {
		this._capacity = values::energyCapacity
		this._receive = values::maxEnergyReceive
		this._extract = values::maxEnergyExtract
		this._initialBatteryLevel = values::initialBatteryLevel
	}

	constructor() : super(Properties().stacksTo(1).rarity(Rarity.EPIC)) {
		_capacity = { Decimal.POSITIVE_INFINITY }
		_receive = { Decimal.POSITIVE_INFINITY }
		_extract = { Decimal.POSITIVE_INFINITY }
		_initialBatteryLevel = { Decimal.POSITIVE_INFINITY }
	}

	override fun isBarVisible(p_150899_: ItemStack): Boolean {
		if (_capacity.invoke().isInfinite)
			return false

		return p_150899_.matteryEnergy != null
	}

	override fun getBarWidth(p_150900_: ItemStack): Int {
		if (_capacity.invoke().isInfinite)
			return 13

		return p_150900_.matteryEnergy?.getBarWidth() ?: super.getBarWidth(p_150900_)
	}

	override fun getBarColor(p_150901_: ItemStack): Int {
		if (_capacity.invoke().isInfinite)
			return 0

		return p_150901_.matteryEnergy?.getBarColor() ?: super.getBarColor(p_150901_)
	}

	init {
		tooltips.itemEnergy()
	}

	private fun capFor(stack: ItemStack): EnergyCapacitorItem {
		return EnergyCapacitorItem(stack, this@BatteryItem.capacity, this@BatteryItem.receive, this@BatteryItem.extract, initialBatteryLevel = this@BatteryItem.initialBatteryLevel)
	}

	override fun registerCapabilities(event: RegisterCapabilitiesEvent) {
		event.registerItem(MatteryCapability.ITEM_ENERGY, { o, _ -> capFor(o) }, this)
		event.registerItem(Capabilities.EnergyStorage.ITEM, { o, _ -> capFor(o) }, this)
	}
}

class CrudeBatteryItem : BatteryItem(ItemsConfig.Batteries.CRUDE) {
	init {
		tooltips.addNormal { itemStack, context, acceptor ->
			val isAndroid = runIfClient(false) { minecraft.player?.matteryPlayer?.isAndroid ?: false }

			if (isAndroid) {
				acceptor(TranslatableComponent("otm.gui.crude_battery.replace_in_world").withStyle(ChatFormatting.GRAY))
				acceptor(TranslatableComponent("otm.gui.crude_battery.replace_in_world_warning").withStyle(ChatFormatting.DARK_RED))
			}
		}
	}

	override fun getUseAnimation(itemStack: ItemStack): UseAnim {
		return UseAnim.BOW
	}

	override fun getUseDuration(itemStack: ItemStack, p_344979_: LivingEntity): Int {
		return 100
	}

	override fun use(level: Level, player: Player, hand: InteractionHand): InteractionResultHolder<ItemStack> {
		if (player.matteryPlayer.isAndroid) {
			player.startUsingItem(hand)
			return InteractionResultHolder.consume(player.getItemInHand(hand))
		}

		return super.use(level, player, hand)
	}

	override fun finishUsingItem(itemStack: ItemStack, level: Level, player: LivingEntity): ItemStack {
		if (player !is Player) return super.finishUsingItem(itemStack, level, player)
		val mattery = player.matteryPlayer ?: return super.finishUsingItem(itemStack, level, player)
		if (!mattery.isAndroid) return super.finishUsingItem(itemStack, level, player)

		if (player is ServerPlayer) {
			if (!mattery.androidEnergy.item.isEmpty) {
				mattery.androidEnergy.item.getCapability(Capabilities.EnergyStorage.ITEM)?.let {
					it.extractEnergy((it.maxEnergyStored * level.random.nextFloat() * .2f).roundToInt(), false)
				}

				mattery.dropBattery()
			}

			val copyStack = itemStack.copy()
			copyStack.count = 1
			mattery.androidEnergy.item = copyStack

			val extraDamageMult = level.random.nextFloat()
			player.hurt(MatteryDamageSource(level.registryAccess().damageType(MDamageTypes.EMP), inflictor = itemStack), 1.5f + extraDamageMult * 3.5f)

			val debuffDuration = 100 + (100 * (1f - extraDamageMult)).roundToInt()
			player.addEffect(MobEffectInstance(MobEffects.BLINDNESS, debuffDuration), player)
			player.addEffect(MobEffectInstance(MobEffects.MOVEMENT_SLOWDOWN, debuffDuration, 1), player)
			player.addEffect(MobEffectInstance(MobEffects.DIG_SLOWDOWN, debuffDuration, 2), player)
			player.addEffect(MobEffectInstance(MobEffects.WEAKNESS, debuffDuration, 2), player)
		}

		if (!player.abilities.instabuild)
			itemStack.shrink(1)

		return itemStack
	}
}
